/*
 *                Copyright (c) Apr 10 19:15:10  MTCC
 *	Author: Michael Thomas
 *	Module: hparse.c
 *	Created: Thu Apr 10 19:15:10 1997
 *	Abstract:
 *	   main parsing routines. Though this module doesn't
 *         parse every entity, it's easy to extend to the
 *         elements which are interesting.
 */

/* Edit History: 
 */

#include "html.h"
#include <time.h>

/*
 * dumb structure to associate names with values
 */



#define ALLOCINC 128

static void err (char * fmt, ...);
static parsetag ();
static int lineno;
static const char * curfile;


static inline void
addc (htelm * p, char ch) 
{
    if (p->len >= p->alen - 1) {
        p->body = realloc (p->body, p->alen + ALLOCINC);
        p->alen += ALLOCINC;
    }
    p->body [p->len++] = ch;
    p->body [p->len] = 0;
}

static inline htelm *
allocelm (int type, int lineno) 
{
    
    htelm * p = calloc (sizeof (htelm), 1);
    p->type = type;
    p->body = malloc (ALLOCINC);
    * p->body = 0;
    p->alen = ALLOCINC;
    p->lineno = lineno;
    return p;
}


static inline int htgetc (FILE * fp, const char ** str) 
{
    int rv;
    if (fp)
        return getc (fp);
    if (** str == 0)
        return EOF;
    rv = ** str;
    * str = * str+1;
    return rv;
}

   
        
static htelm *
htread (htelm * head, htelm **pend, FILE * fp, const char * str, int ishtml) 
{
    int ch, rv;
    htelm * h, * new;
    int inquote = 0;

    /* always splice in a new node */
    lineno = 1, 
    h = allocelm (H_TXT, lineno);    
    if (head == 0)
        head = h;
    else {
        if (head->flink) {
            h->flink = head->flink;
            head->flink->blink = h;
        }
        head->flink = h;        
        h->blink = head;
        h->ulink = head->ulink;
    }
    while ((ch = htgetc (fp, &str)) != EOF) {
        /* fudge the character to be something uninteresting
           if we're not parsing html */
        switch (ishtml ? ch : ' ') {
            case '<':
                if (inquote) {
                    addc (h, ch);
                    break;
                }
                /* start a new tag, only if necessary */
                if (h->type != H_TXT) {
                    err ("spurious open tag around '%.50s'", h->blink->body);
                    goto errcleanup;                    
                }
                if (h->type != H_TXT || h->len != 0) {
                    new = allocelm (H_MISCTAG, lineno);
                    /* inherit uplink */
                    new->ulink = h->ulink;
                    if (h->flink) {
                        new->flink = h->flink;
                        new->flink->blink = new;
                    }                    
                    h->flink = new;
                    new->blink = h;
                    h = new;
                } else
                    h->type = H_MISCTAG;                
                addc (h, ch);                
                break;                
            case '>':
                addc (h, ch);
                if (inquote)
                    break;
                if (h->type != H_MISCTAG) {
                    if (h->blink)
                        err ("spurious '>' around '%.50s'", h->blink->body);
                    else
                        err ("spurious '>'");
                    break;
                }
                /* parse the tag, see if we should push/pop context */
                new = allocelm (H_TXT, lineno);
                if ((rv = parsetag (&h))) {
                    if (rv > 0) {
                        /* start a new list off of slink */
                        h->slink = new;
                        new->ulink = h;
                    } else {
                        if (rv == -2) {
                            err ("unbalanced tag %.50s", h->body);
                            goto errcleanup;
                        }
                        if (h->ulink == 0) {
                            err ("unexpected close tag %.50s", h->body);
                            goto errcleanup;
                        }
                        /* pop */
                        h = h->ulink;
                        if (h->flink) {
                            new->flink = h->flink;
                            new->flink->blink = new;
                        }
                        h->flink = new;
                        new->blink = h;
                        new->ulink = h->ulink;
                    }
                } else {
                    if (h->flink) {
                        new->flink = h->flink;
                        new->flink->blink = new;
                    }                    
                    h->flink = new;
                    new->blink = h;
                    new->ulink = h->ulink;
                }
                h = new;                
                break;
            default:
                if (ch == '\n' || ch == '\f') {
                    lineno++;
                    /* seal off the current tag */
                    if (h->type == H_TXT) {   
                        new = allocelm (H_TXT, lineno);
                        if (h->flink) {
                            new->flink = h->flink;
                            new->flink->blink = new;
                        }                    
                        h->flink = new;
                        new->blink = h;
                        new->ulink = h->ulink;
                        h = new;
                    }
                }
                if (h->type && ch == '\"')
                    inquote ^= 1;
                addc (h, ch);
                break;
        }
    }
  errcleanup:    
#if 0    
    if (h->ulink != head->ulink)
        err ("eof reached with unclosed tag %.50s\n", h->ulink->body);
#endif    
    if (pend)
        *pend = h;
    return head;
}

htelm * htpreadstr (htelm * head, htelm **pend, const char * str)
{
    htelm * h;
    int svline = lineno;
    const char * svfile = curfile;
    
    curfile = str;
    h = htread (head, pend, 0, str, 1);
    curfile = svfile;
    lineno = svline;
    return h;
}

/* read a file to the current file */

htelm * htpreadfile (htelm * head, htelm **pend, const char * file) 
{
    FILE * fp = fopen (file, "r");
    htelm * h;
    int svline = lineno;
    const char * svfile = curfile;
    
    if (fp == NULL)
        return 0;
    curfile = file;
    h = htread (head, pend, fp, 0, 1);
    curfile = svfile;
    lineno = svline;
    fclose (fp);
    return h;
}
/*
 * read a file, but just consider it text
 */

htelm * htpreadfiletext (htelm * head, htelm **pend, const char * file) 
{
    FILE * fp = fopen (file, "r");
    htelm * h;
    int svline = lineno;
    const char * svfile = curfile;
    
    if (fp == NULL)
        return 0;
    curfile = file;
    h = htread (head, pend, fp, 0, 0);
    curfile = svfile;
    lineno = svline;
    fclose (fp);
    return h;
}

/* read the given URL  */

htelm * htpreadurl (htelm * head, htelm **pend, const char * url, int * psts, char * type) 
{
    FILE * fp;
    htelm * h = 0;
    int svline = lineno;
    const char * svfile = curfile;
    char content [256];
    
    if (type == 0)
        type = content;
    * type = 0;    
    if ((fp = fopenurl ((char *) url, psts, type)) == 0)
        return 0;
    
    curfile = url;
    /* if specifically html parse as such */
    if (! strcasecmp (type, "text/html"))
        h = htread (head, pend, fp, 0, 1);
    else if (! strncasecmp (type, "text/", 5))
        /* otherwise, if some other type of text, turn off html part */
        h = htread (head, pend, fp, 0, 0);
    /* else ignore it altogether. */
    curfile = svfile;
    lineno = svline;
    fclose (fp);
    return h;
}




static int dobranch (), doinput (), doinc (), doset ();

struct varval tags [] = {    { "form", H_FORM, dobranch },
                             { "input", H_INPUT, doinput },
                             { "option", H_OPTION, 0 },
                             { "select", H_SELECT, dobranch },
                             { "textarea", H_TXTAREA, dobranch },
                             { "a", H_ANCHOR, 0 },                             
                             { "$table", H_TABLEDEF, dobranch },
                             { "table", H_TABLE, dobranch },
                             { "th",   H_TBHDR, dobranch },
                             { "tr",   H_TBROW, dobranch },
                             { "td",   H_TBCELL, dobranch },
                             { "$mark", H_MARKER, 0 },
                             { "$date", H_DATE, 0 },
                             { "$include", H_INC, doinc },
                             { "$if", H_IF, dobranch },
                             { "$else", H_ELSE, 0 },
                             { "$set", H_SET, 0 }, 
                             { 0 }
};


/*
 * tokenize a string following 
 */

static struct token *
tokenize (char * txt) 
{
    struct token * head = 0, * t, * prev = 0;
    char word [1024];
    int len = 0, goteq = 0;

    if (* txt++ != '<' || * txt == '!')
        return 0;
    head = t = calloc (sizeof (* t), 1);
    t->flags = TK_TAG;
    for (; * txt; txt++) {
        if (* txt == '/' && (t->flags & TK_TAG)) {
            t->flags |= TK_ISEND;
            continue;
        }
        switch (* txt) {
            case '\"':
                txt++;
                /* behave like normal, but strip the quotes */
                while (* txt && * txt != '\"')
                    if (len < 1023)
                        word [len++] = * txt++;
                    else
                        txt++;
                t->flags |= TK_HASQ;
                word [len] = 0;
                if (* txt == 0)
                    err ("unclosed quote in text '%.50s'", word);
                break;
            case '=':
                if (len == 0)
                    goteq = 1;
                else
                    goteq = 2;                
            default:
                if ((isspace (* txt) || * txt == '=' || *txt == '>')) {
                    if (len == 0 && (t->flags & TK_HASQ) == 0)
                        break;
                    /* seal off this tag */
                    if (prev)
                        prev->flink = t;
                    t->name = strdup (word);
                    if (goteq == 1){
                        t->flags |= TK_VAL;
                        t->flags &= ~ TK_VAR;
                    }
                    prev = t;
                    t = calloc (sizeof (* t), 1);
                    t->flags = goteq == 2 ? TK_VAL : TK_VAR;
                    len = 0;
                    goteq = 0;                    
                } else {
                    if (len < 1023) {
                        word [len++] = * txt;
                        word [len] = 0;
                    }
                }                
                break;
        }
    }
    if (head->name == 0) {
        free (head);
        return 0;
    }
    return head;
}


/*
 * get the n'th word in the token list
 * observed.
 */

struct token *
htkgetword (struct token * tk, int n) 
{
    while (n-- > 0) {
        if (tk->flink == 0)
            return 0;
        tk = tk->flink;
    }
    return tk;
}

/*
 * get the value of var in the given token list
 */

char *
htkgetval (struct token * tk, char * var) 
{
    for (; tk; tk = tk->flink) {
        if ((tk->flags & TK_VAR) && ! strcasecmp (tk->name, var))
            if (tk->flink) {
                tk = tk->flink;
                if (tk->flags & TK_VAL)
                    return tk->name;
            } else
                return "";      /* debatable: found the var, but no val */
    }
    /* this is clearly an error. */
    return 0;
}

struct token * 
htkckvar (struct token * tk, char * var) 
{
    for (; tk; tk = tk->flink) {
        if ((tk->flags & TK_VAR) && ! strcasecmp (tk->name, var))
            return tk;
    }
    /* this is clearly an error. */
    return 0;
}

        
 

/*
 * decompose a tag. The return value is -1..1 which
 * means pop, donothing, push respectively.
 */
static int
parsetag (htelm ** ph) 
{
    struct varval * t;
    htelm * h = * ph;
    
    if ((h->tk = tokenize (h->body)) == 0) {
        /* boring for some reason, bail now */
        return 0;
    }
    
    /* should bsearch the array, but linear is ok if it's
       small */
    for (t = tags; t->name; t++) {
        if (! strncasecmp (t->name, h->tk->name, strlen (t->name))) {
            h->type = t->type;
            if (t->fn == 0)
                return 0;
            return t->fn (ph);
        }
    }
    return 0;
}

static int
dobranch (htelm ** ph) 
{
    htelm * h = * ph;
    
    int rv = h->tk->flags & TK_ISEND ? -1 : 1;
    /* check up to see if the that we're matching the right thing. */
    if (rv < 0) {
        if (h->ulink == 0 || h->ulink->type != h->type) {
            return -2;
        }
        
    }
    return rv;
}

        

static struct varval intype [] = { {"radio", H_RADIO},
                                 {"checkbox", H_CHECK },
                                 { "submit", H_SUBMIT},
                                 {"reset", H_RESET },
                                 {"text", H_INPUT },
                                 {"password", H_PASSWORD },
                                 { 0 }
};
  

static int
doinput (htelm ** ph) 
{
    char * p = htkgetval ((*ph)->tk, "type");
    struct varval * w;
    
    if (p) 
        for (w = intype; w->name; w++)
            if (! strcasecmp (w->name, p)) {
                (*ph)->type = w->type;
                return 0;
            }
    return 0;
}



static int
doinc (htelm ** ph) 
{
    /* no include? bzzt */
    if ((*ph)->tk->flink == 0)
        return 0;
    /* include a file at h */
    htpreadfile (* ph, ph, (*ph)->tk->flink->name);
    return 0;
}

static int
doset (htelm * h) 
{
    /* no include? bzzt */
    if (h->tk->flink == 0)
        return 0;
    hputenv (h->tk->flink->name,
             h->tk->flink->flink ? h->tk->flink->flink->name : "");
#if 0    
    printf ("set: %s=%s", h->tk->flink->name,
             h->tk->flink->flink ? h->tk->flink->flink->name : "");
#endif    
    return 0;
}

static int silence = 1;


static void err (char * fmt, ...) 
{
    va_list ap;
    char bf [512];
    if (silence)
        return;
    VA_START (ap, fmt);
    vsprintf (bf, fmt, ap);
    va_end (ap);
    fprintf (stderr, "htparse: @%d in file/text = '%-.50s'\n", lineno, curfile);
    fprintf (stderr, "--> %s\n", bf);    
}

void hterrors (int on) 
{
    silence = ! on;
}




/*
 * release the tree starting at h
 */

void htprelease (htelm * h) 
{
    htelm * p;
    struct token * tk;
    
    while ((p = h)) {
        if ((h->slink))
            htprelease (h->slink);
        h = h->flink;
        if ((p->body))
            free (p->body);
        while ((tk = p->tk)) {
            p->tk = p->tk->flink;
            if (tk->name)
                free (tk->name);
            free (tk);
        }
        free (p);
    }
}

        


void htpdumpfile (htelm * h) 
{
    _htpdumpfile (h, -1, printf);
    fflush (stdout);
}



void _htpdumpfile (htelm * h, int endtag, int ((* dumper) (const char * fmt, ...))) 
{
    htelm * p;
    while (h) {
        if (h->type == endtag)
            break;
        /* text and unknown goobers get emitted as is */
        if (h->type == H_TXT || h->tk == 0) 
            dumper ("%s", h->body);
        else if (h->type < H_PSSTART) {
            struct token * tk;

            dumper ("<");
            for (tk = h->tk; tk; tk = tk->flink) {
                if (tk->flags & TK_ISEND)
                    dumper ("/%s", tk->name);
                else if (tk->flags & TK_VAL) {
                    if (tk->flags & TK_HASQ)
                        dumper ("=\"%s\"", tk->name);
                    else {
                        /* variable expansion? */
                        if (* tk->name == '$') {
                            char * v = hgetenv (tk->name+1);
                            if (v)
                                dumper ("=\"%s\"", v);
                            else
                                dumper ("=???");
                        } else                        
                            dumper ("=%s", tk->name);
                    }
                    
                } else {
                    dumper ("%s", tk->name);
                }
                if (tk->flink && (tk->flink->flags & TK_VAL) == 0)
                    dumper (" ");                
            }
            dumper (">");
        } else {
            switch (h->type) {
                case H_DATE: 
                {
                    long tm;
                    time (&tm);
                    dumper ("%s", ctime (&tm));
                    break;
                }
                case H_TABLEDEF:
                    /* don't go down here. */
                    h = h->flink;
                    continue;
                case H_IF:
                    if ((h->tk->flags & TK_ISEND) != 0)
                        break;

                    if (hcond (h) == 0) {
                        /* search the next level down foran else */
                        for (p = h->slink; p != 0; p = p->flink)
                            if (p->type == H_ELSE)
                                break;
                        if (p != 0)
                            _htpdumpfile (p->flink, H_NOTAG, dumper);
                        h = h->flink;
                    } else {
                        _htpdumpfile (h->slink, H_ELSE, dumper);
                        h = h->flink;
                    }
                    /* note that we don't exec the bottom 
                       of this loop. */
                    continue;                        
/*                    
                    if (h->tk->flags & TK_ISEND)
                        break;
                    if (hcond (h) == 0)
                        h = h->flink;
                    break;
                    */                    
                case H_SET:
                    doset (h);
                    break;
                default:
                    break;                    
            }
        }
        /* do not consider endtags of inferior branches */
        if (h == 0)
            return;
        if (h->slink)
            _htpdumpfile (h->slink, H_NOTAG, dumper);        
        h = h->flink;
    }
}


/*
 * post mortem tree dumper
 */

void htpdumptree (htelm * h, int lvl) 
{

    if (lvl == 0) {
        hterase ();
        preHTML ();
        startHTML ("htdump", "");
    }
    
    bq (1);
    while (h) {        
        htprintf ("#%d ", h->lineno);
        if (h->type == H_TXT || h->tk == 0) {
            htprintf ("txt: '%-20.20s%s'", h->body,
                    strlen (h->body) > 10 ? "..." : "");
        } else if (h->type < H_PSSTART) {
            struct token * tk;
            htprintf ("tag: %d &lt", h->type);        
            for (tk = h->tk; tk; tk = tk->flink) {
                if (tk->flags & TK_ISEND)
                    htprintf ("/%s", tk->name);
                else if (tk->flags & TK_VAL) {
                    if (tk->flags & TK_HASQ)
                        htprintf ("=\"%s\"", tk->name);
                    else
                        htprintf ("=%s", tk->name);
                } else {
                    htprintf ("%s", tk->name);
                }
                if (tk->flink && (tk->flink->flags & TK_VAL) == 0)
                    htprintf (" ");
            }
            htprintf ("&gt");
        } else {
            switch (h->type) {
                case H_DATE: 
                {
                    long tm;
                    time (&tm);
                    htprintf ("meta: %s", ctime (&tm));
                    break;
                }
                default:
                    htprintf ("meta: %s", h->body);
                    break;
            }
        }
        br (1);
        if (h->slink) {
            htpdumptree (h->slink, 1);
        }
        h = h->flink;
    }
    bq (-1);
    if (lvl == 0) {
        endHTML ();
        exit (0);
    }
}



/*
 * compile the internal representation into a bunch
 * of C structures suitable for use at run time.
 * (ie Casting policy into Concrete)
 */

int htpcompile (htelm * h) 
{
    return 0;
}


