/*
 *                 Copyright (c) Jun 20 17:47:54  MTCC
 *	Author: Michael Thomas
 *	Module: util.c
 *	Created: Fri Jun 20 17:47:54 1997
 *	Abstract:
 *	   utility functions for the mtccbase
 */

/* Edit History: 
 */

#include "db.h"

static boolean bigendian;


writefully (int fd, u8 * bf, int len) 
{
    int rv;
    int slen = len;
    
    do {
        rv = write (fd, bf, len);
        if (rv <= 0)
            return rv;
        len -= rv;
        bf += rv;
    } while (len > 0);
    return slen;
}

    

readfully (int fd, u8 * bf, int len) 
{
    int rv;
    int slen = len;

    do {
        rv = read (fd, bf, len);
        if (rv <= 0)
            return rv;
        len -= rv;
        bf += rv;
    } while (len > 0);
    return slen;
}


    

/* endianness independent functions */

/* figure out whether we're big or little endian */

setendian () 
{
    short bf;
    char * p;

    p = (char *) &bf;
    p [0] = 1;
    p [1] = 2;

    if (bf == 0x12)
        bigendian = true;
    else
        bigendian = false;
}

void 
htovl (u32 val, u8 * bf) 
{
    u8 * p = (u8 *) &val;
    if (bigendian) {
        bf [0] = p [3];
        bf [1] = p [2];
        bf [2] = p [1];
        bf [3] = p [0];
    } else {
        bf [0] = p [0];
        bf [1] = p [1];
        bf [2] = p [2];
        bf [3] = p [3];
    }
}

u32
vtohl (u8 * bf) 
{
    u32 rv;
    u8 * p = (u8 *) &rv;
    
    if (bigendian) {
        p [0] = bf [3];
        p [1] = bf [2];
        p [2] = bf [1];
        p [3] = bf [0];
    } else {
        p [0] = bf [0];
        p [1] = bf [1];
        p [2] = bf [2];
        p [3] = bf [3];
    }
    return rv;
}


htovd (double val, u8 * bf) 
{
    u8 * p = (u8 *) &val;
    if (bigendian) {
        bf [0] = p [7];
        bf [1] = p [6];
        bf [2] = p [5];
        bf [3] = p [4];
        bf [4] = p [3];
        bf [5] = p [2];
        bf [6] = p [1];
        bf [7] = p [0];        
    } else {
        bf [0] = p [0];
        bf [1] = p [1];
        bf [2] = p [2];
        bf [3] = p [3];        
        bf [4] = p [4];        
        bf [5] = p [5];        
        bf [6] = p [6];        
        bf [7] = p [7];
    }
}


double
vtohd (u8 * bf) 
{
    double rv;
    u8 * p = (u8 *) &rv;
    
    if (bigendian) {
        p [0] = bf [7];
        p [1] = bf [6];
        p [2] = bf [5];
        p [3] = bf [4];
        p [4] = bf [3];
        p [5] = bf [2];
        p [6] = bf [1];
        p [7] = bf [0];
    } else {
        p [0] = bf [0];
        p [1] = bf [1];
        p [2] = bf [2];
        p [3] = bf [3];
        p [4] = bf [4];
        p [5] = bf [5];
        p [6] = bf [6];
        p [7] = bf [7];        
    }
    return rv;
}


/* pack a record into database format */
/* note the length includes the record header */

u8 *
pack (mtccdb * db, hrec * r, int * plen) 
{
    hrecelm * h, * local;    
    int len = 0;
    u8 * p, * bf;

    /* first, find out the overall length of the
       record. this is sort of gross, but saves the
       likelihood of realloc type copy thrashing */
    h = hrfirst (db->cleanrec);
    
    while (h != null) {
        /* do a bit of sanity checking */
        if ((local = hrgeth (r, h->var)) == null)
            local = h;

        if (h->fmt != local->fmt) {
            fprintf (stderr, "illegal type clash for %s\n", h->var);
            return null;
        }
        
        switch (local->fmt) {
            case HRF_TEXT:
                len += strlen (local->val)+1;
                break;
            case HRF_CHAR: 
                /* always use the fullrec subfmt: it's reliable */
                len += h->subfmt;
                break;
            case HRF_BINARY:
                len += local->subfmt+4;
                break;                
            case HRF_INT:
                len += 4;
                break;
            case HRF_DOUBLE:
                len += 8;
                break;
        }
        h = hrnext (db->cleanrec);
    }
    len += 4;
    
    /* pad len to BLKSIZE */    
    if ((len & BLKMASK) != 0)
        len += BLKSIZE - (len & BLKMASK);        
    * plen = len;
    p = bf = malloc (len);
    
    htovl (len, bf);
    p += 4;
    
    h = hrfirst (db->cleanrec);
    /* now, actually pack the record */
    while (h != null) {
        /* apply defrec default if necessary */
        if ((local = hrgeth (r, h->var)) == null)
            local = h;
        
        switch (local->fmt) {
            case HRF_TEXT:
                len = strlen (local->val)+1;
                memcpy (p, local->val, len);
                break;
            case HRF_CHAR: 
                /* always use the fullrec subfmt: it's reliable */
                if (h->subfmt)                
                    memcpy (p, local->val, h->subfmt);
                len = h->subfmt;
                break;
            case HRF_BINARY:
                htovl (local->subfmt+4, p);
                if (local->subfmt)
                    memcpy (p+4, local->val, local->subfmt);
                len = local->subfmt+4;
                break;                
            case HRF_INT:
                htovl (local->cvtval.ival, p);
                len = 4;
                break;
            case HRF_DOUBLE:
                htovd (local->cvtval.dval, p);
                len = 8;
                break;
            default:
                len = 0;
        }
        p += len;
        h = hrnext (db->cleanrec);
    }
    return bf;        
}

/* unpack a record into an hrec */

unpack (mtccdb * db, u8 * bf, hrec * r) 
{
    hrecelm * h;
    double dval;
    int ival, len, vis;
    
    /* use fullrec since this is an entire database payload */
    h = hrfirst (db->fullrec);
    while (h != null) {
        vis = 0;
        if (hrisvis (r, h->var))
            vis = 1;
        switch (h->fmt) {
            case HRF_TEXT:
                len = strlen (bf)+1;
                if (vis)
                    hradd (r, h->var, bf);
                break;
            case HRF_CHAR:
                len = h->subfmt;
                if (vis)
                    hraddbf (r, h->var, bf, len);
                break;
            case HRF_BINARY:
                len = vtohl (bf);                
                if (vis)
                    hraddbin (r, h->var, bf+4, len-4);
                break;                
            case HRF_INT:
                ival = vtohl (bf);
                len = 4;
                if (vis)
                    hraddint (r, h->var, ival);
                break;
            case HRF_DOUBLE:
                dval = vtohd (bf);
                len = 8;
                if (vis)
                    hradddouble (r, h->var, dval);   
                break;
        }
        bf += len;
        h = hrnext (db->fullrec);
    }
}

        
